from .agent import JAImsAgent
from .default_tool_manager import JAImsDefaultToolManager
from .default_history_manager import (
    JAImsDefaultHistoryManager,
    JAImsLastNHistoryOptimizer,
)
from .entities import (
    JAImsMessage,
    JAImsImageContent,
    JAImsContentType,
    JAImsMessageRole,
    JAImsToolResponse,
    JAImsFunctionTool,
    JAImsMaxConsecutiveFunctionCallsExceeded,
    JAImsFunctionToolDescriptor,
    JAImsParamDescriptor,
    JAImsJsonSchemaType,
    JAImsToolCall,
)
from .interfaces import JAImsToolManager, JAImsHistoryManager, JAImsLLMInterface



GEMINI_1_PRO = "gemini-1.0-pro"
GEMINI_1_PRO_LATEST = "gemini-1.0-pro-latest"
GEMINI_1_PRO_001 = "gemini-1.0-pro-001"
GEMINI_1_PRO_VISION = "gemini-1.0-pro-vision"
GEMINI_1_PRO_VISION_LATEST = "gemini-1.0-pro-vision-latest"
GEMINI_1_5_FLASH_LATEST = "gemini-1.5-flash-latest"
GEMINI_1_5_PRO = "gemini-1.5-pro"
GEMINI_1_5_PRO_LATEST = "gemini-1.5-pro-latest"



from __future__ import annotations

# Enum class over all Json Types
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Union
from PIL import Image


# -------------------------
# LLM Messaging Abstraction
# -------------------------


class JAImsToolCall:
    def __init__(self, id: str, tool_name: str, tool_args: Optional[Dict[str, Any]]):
        self.id = id
        self.tool_name = tool_name
        self.tool_args = tool_args


class JAImsToolResponse:
    def __init__(self, tool_call_id: str, tool_name: str, response: Any):
        self.tool_call_id = tool_call_id
        self.tool_name = tool_name
        self.response = response


JAImsImageContentType = Union[str, Image.Image]


class JAImsImageContent:
    def __init__(self, image: JAImsImageContentType):
        self.image = image


JAImsContentType = Union[JAImsImageContent, str]


class JAImsMessageRole(Enum):
    USER = "user"
    ASSISTANT = "assistant"
    TOOL = "tool"
    SYSTEM = "system"


class JAImsMessage:
    def __init__(
        self,
        role: JAImsMessageRole,
        contents: Optional[List[JAImsContentType]] = None,
        name: Optional[str] = None,
        tool_calls: Optional[List[JAImsToolCall]] = None,
        tool_response: Optional[JAImsToolResponse] = None,
        raw: Optional[Any] = None,
    ):
        self.role = role
        self.contents = contents
        self.name = name
        self.tool_calls = tool_calls
        self.tool_response = tool_response
        self.raw = raw

    def get_text(self) -> Optional[str]:
        if self.contents is None:
            return None

        all_text = ""
        for content in self.contents:
            if isinstance(content, str):
                all_text += content

        return all_text or None

    # -------------------------
    # convenience constructors
    # -------------------------

    @staticmethod
    def user_message(
        text: str,
        raw: Optional[Any] = None,
    ) -> JAImsMessage:
        return JAImsMessage(
            role=JAImsMessageRole.USER,
            contents=[text],
            raw=raw,
        )

    @staticmethod
    def assistant_message(text: str, raw: Optional[Any] = None) -> JAImsMessage:
        return JAImsMessage(
            role=JAImsMessageRole.ASSISTANT,
            contents=[text],
            raw=raw,
        )

    @staticmethod
    def system_message(text: str, raw: Optional[Any] = None) -> JAImsMessage:
        return JAImsMessage(
            role=JAImsMessageRole.SYSTEM,
            contents=[text],
            raw=raw,
        )

    @staticmethod
    def tool_response_message(
        tool_call_id: str, tool_name: str, response: Any, raw: Optional[Any] = None
    ) -> JAImsMessage:
        return JAImsMessage(
            role=JAImsMessageRole.TOOL,
            tool_response=JAImsToolResponse(tool_call_id, tool_name, response),
            raw=raw,
        )

    @staticmethod
    def tool_call_message(
        tool_calls: List[JAImsToolCall], raw: Optional[Any] = None
    ) -> JAImsMessage:
        return JAImsMessage(
            role=JAImsMessageRole.ASSISTANT,
            tool_calls=tool_calls,
            raw=raw,
        )


class JAImsStreamingMessage:
    def __init__(self, message: JAImsMessage, textDelta: Optional[str] = None):
        self.message = message
        self.textDelta = textDelta


# ------------------------------------------
# Params, tool and function handling classes
# ------------------------------------------


class JAImsJsonSchemaType(Enum):
    STRING = "string"
    NUMBER = "number"
    OBJECT = "object"
    ARRAY = "array"
    BOOLEAN = "boolean"
    NULL = "null"


class JAImsParamDescriptor:
    """
    Describes a parameter to be used in the OPENAI API.

    Attributes
    ----------
        name : str
            the parameter name
        description : str
            the parameter description
        json_type : JsonType:
            the parameter json type
        attributes_params_descriptors : list of JAImsParamDescriptor
            the list of parameters descriptors for the attributes of the parameter
            in case the parameter is an object, defualts to None
        array_type_descriptors : list of JAImsParamDescriptor
            the parameter descriptors for the array type in case the parameter is an array, defaults to None
        enum_values:
            the list of values in case the parameter is an enum, defaults to None
        required : bool
            whether the parameter is required or not, defaults to True

    """

    def __init__(
        self,
        name: str,
        description: str,
        json_type: JAImsJsonSchemaType,
        attributes_params_descriptors: Optional[List[JAImsParamDescriptor]] = None,
        array_type_descriptors: Optional[List[JAImsParamDescriptor]] = None,
        array_type_any_valid: bool = True,
        enum_values: Optional[List[Any]] = None,
        required: bool = True,
    ):
        self.name = name
        self.description = description
        self.json_type = json_type
        self.attributes_params_descriptors = attributes_params_descriptors
        self.array_type_descriptors = array_type_descriptors
        self.array_type_any_valid = array_type_any_valid
        self.enum_values = enum_values
        self.required = required

    def get_jsonapi_schema(self) -> Dict[str, Any]:
        """
        Returns the jsonapi schema for the parameter.
        """
        schema: dict[str, Any] = {
            "type": self.json_type.value,
            "description": self.description,
        }

        if (
            self.json_type == JAImsJsonSchemaType.OBJECT
            and self.attributes_params_descriptors
        ):
            schema["properties"] = {}
            schema["required"] = []
            for param in self.attributes_params_descriptors:
                schema["properties"][param.name] = param.get_jsonapi_schema()
                if param.required:
                    schema["required"].append(param.name)

        if self.json_type == JAImsJsonSchemaType.ARRAY and self.array_type_descriptors:
            items_schema = [
                desc.get_jsonapi_schema() for desc in self.array_type_descriptors
            ]
            if self.array_type_any_valid:
                schema["items"] = {"anyOf": items_schema}
            else:
                schema["items"] = [items_schema]

        if self.enum_values:
            schema["enum"] = self.enum_values

        return schema


class JAImsFunctionToolDescriptor:
    """
    Describes a tool to be used in the OPENAI API. Supports only function tool for now.

    Attributes
    ----------
        name : str
            the tool name
        description : str
            the tool description
        params_descriptors: List[JAImsParamDescriptor]
            the list of parameters descriptors
    """

    def __init__(
        self,
        name: str,
        description: str,
        params_descriptors: List[JAImsParamDescriptor],
    ):
        self.name = name
        self.description = description
        self.params_descriptors = params_descriptors

    def get_jsonapi_schema(self) -> Dict[str, Any]:
        """
        Returns the jsonapi schema for function.
        """
        schema = {
            "type": "object",
            "properties": {},
            "required": [],
        }

        for param in self.params_descriptors:
            schema["properties"][param.name] = param.get_jsonapi_schema()
            if param.required:
                schema["required"].append(param.name)

        return schema


class JAImsFunctionTool:
    """
    Wraps a function tool used by the LLM along with a function to be called locally when
    the tool is invoked by the LLM.
    You may subclass this class to implement your own function wrapper behavior by overriding the call method.


    Attributes
    ----------
        function : Callable[..., Any]
            The function to be called when the tool is invoked, defaults to None.
            When None, the tool call pass None to the agent as a result.
        to_openai_function_tool : JAImsFunctionToolDescriptor
            The tool descriptor, contains the markup information that will be used to be passed
            as a tool invocation dictionary to the LLM.

    Methods
    -------
        call(params: Dict[str, Any]) -> Any
            Calls the wrapped function with the given parameters, if the function is not None.
            Returns None otherwise.
    """

    def __init__(
        self,
        function_tool_descriptor: JAImsFunctionToolDescriptor,
        function: Optional[Callable[..., Any]] = None,
    ):
        self.function = function
        self.function_tool = function_tool_descriptor

    def call(self, params: Optional[Dict[str, Any]]) -> Any:
        """
        Calls the wrapped function with the passed parameters if the function is not None.
        Returns None otherwise.

        Parameters
        ----------
            params : dict
                the parameters passed to the wrapped function
        """

        return self.function(**params) if params and self.function else None


# ----------
# exceptions
# ----------


class JAImsMaxConsecutiveFunctionCallsExceeded(Exception):
    """
    Exception raised when the maximum number of consecutive function calls is exceeded.

    Attributes:
        max_consecutive_calls -- maximum number of consecutive calls allowed
    """

    def __init__(self, max_consecutive_calls):
        message = f"Max consecutive function calls exceeded: {max_consecutive_calls}"
        super().__init__(message)


class JAImsUnexpectedFunctionCall(Exception):
    """
    Exception raised when an unexpected function call occurs.

    Attributes:
        func_name -- name of the unexpected function
    """

    def __init__(self, func_name):
        message = f"Unexpected function call: {func_name}"
        super().__init__(message)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .agent import JAImsAgent

from typing import List
from jaims.entities import (
    JAImsFunctionTool,
    JAImsUnexpectedFunctionCall,
    JAImsToolCall,
    JAImsMessage,
)

from jaims.interfaces import JAImsToolManager


class JAImsDefaultToolManager(JAImsToolManager):

    def handle_tool_calls(
        self,
        agent: JAImsAgent,
        tool_calls: List[JAImsToolCall],
        tools: List[JAImsFunctionTool],
    ) -> List[JAImsMessage]:
        results = []
        for fc in tool_calls:
            function_name = fc.tool_name
            tool_wrapper = next(
                (f for f in tools if f.function_tool.name == function_name),
                None,
            )
            if not tool_wrapper:
                raise JAImsUnexpectedFunctionCall(function_name)

            fc_result = tool_wrapper.call(fc.tool_args)
            results.append(
                JAImsMessage.tool_response_message(
                    tool_call_id=fc.id,
                    tool_name=fc.tool_name,
                    response=fc_result,
                )
            )

        return results



import os
from ...interfaces import JAImsLLMInterface, JAImsHistoryManager, JAImsToolManager
from ...entities import (
    JAImsMessage,
    JAImsFunctionTool,
    JAImsToolCall,
    JAImsMessageRole,
    JAImsContentType,
    JAImsImageContent,
    JAImsStreamingMessage,
)
from ...agent import JAImsAgent
from ..shared.image_utilities import image_to_bytes

from typing import Iterable, List, Generator, Optional
from PIL import Image


from google.generativeai import GenerativeModel
from google.generativeai.types import content_types
from google.generativeai.types import generation_types
import google.ai.generativelanguage as glm


class JAImsGoogleGenerativeAIAdapter(JAImsLLMInterface):
    def __init__(
        self,
        model: str,
        api_key: Optional[str],
        generation_config: Optional[generation_types.GenerationConfigType],
        tool_config: Optional[content_types.ToolConfigType],
    ):
        self.api_key = api_key or os.getenv("GOOGLE_API_KEY")
        if not self.api_key:
            raise Exception("GOOGLE_API_KEY not provided.")
        self.model = model
        self.generation_config = generation_config
        self.tool_config = tool_config

    def call(
        self, messages: List[JAImsMessage], tools: List[JAImsFunctionTool]
    ) -> JAImsMessage:
        response = self.__get_gemini_response(messages, tools)
        assert isinstance(response, generation_types.GenerateContentResponse)
        return self.__gemini_to_jaims_message(response)

    def call_streaming(
        self, messages: List[JAImsMessage], tools: List[JAImsFunctionTool]
    ) -> Generator[JAImsStreamingMessage, None, None]:
        response = self.__get_gemini_response(messages, tools, stream=True)
        assert isinstance(response, Iterable)
        for r in response:
            if r.candidates:
                message = self.__gemini_to_jaims_message(r)
                yield JAImsStreamingMessage(
                    textDelta=message.get_text() or None,
                    message=message,
                )

    def __jaims_role_to_gemini(self, role: JAImsMessageRole) -> str:
        if role == JAImsMessageRole.USER:
            return "user"
        else:
            return "model"

    def __jaims_tools_to_gemini(
        self, jaims_tools: List[JAImsFunctionTool]
    ) -> List[content_types.Tool]:
        function_declarations = []
        for jaims_tool in jaims_tools:
            function_declarations.append(
                content_types.FunctionDeclaration(
                    name=jaims_tool.function_tool.name,
                    description=jaims_tool.function_tool.description,
                    parameters=jaims_tool.function_tool.get_jsonapi_schema(),
                ),
            )

        return [content_types.Tool(function_declarations=function_declarations)]

    def __jaims_messages_to_gemini(
        self, jaims_messages: List[JAImsMessage]
    ) -> tuple[List[str], List[content_types.ContentsType]]:
        gemini_messages = []
        system_instruction = []

        for jaims_message in jaims_messages:
            if jaims_message.role == JAImsMessageRole.SYSTEM:
                system_instruction.append(jaims_message.get_text())
                continue

            role = self.__jaims_role_to_gemini(jaims_message.role)
            parts = []

            if jaims_message.contents:
                for content in jaims_message.contents:
                    if isinstance(content, str):
                        parts.append(glm.Part(text=content))
                    if isinstance(content, JAImsImageContent):
                        if isinstance(content.image, str):
                            raise ValueError("Image URL not supported")
                        elif isinstance(content.image, Image.Image):
                            mime, data = image_to_bytes(content.image)
                            parts.append(
                                glm.Part(
                                    inline_data=glm.Blob(mime_type=mime, data=data)
                                )
                            )
                        else:
                            raise ValueError("Invalid image content type")

            if jaims_message.tool_response:
                parts.append(
                    glm.Part(
                        function_response=glm.FunctionResponse(
                            name=jaims_message.tool_response.tool_name,
                            response={"result": jaims_message.tool_response.response},
                        )
                    )
                )

            if jaims_message.tool_calls:
                for tool_call in jaims_message.tool_calls:
                    parts.append(
                        glm.Part(
                            function_call=glm.FunctionCall(
                                name=tool_call.tool_name,
                                args=tool_call.tool_args,
                            )
                        )
                    )

            gemini_messages.append(
                glm.Content(
                    role=role,
                    parts=parts,
                )
            )

        return system_instruction, gemini_messages

    def __gemini_to_jaims_message(
        self, response: generation_types.GenerateContentResponse
    ) -> JAImsMessage:
        assert response.candidates is not None

        role = JAImsMessageRole.ASSISTANT
        tool_calls = []
        contents = []
        for part in response.parts:
            if fn := part.function_call:
                args = {}
                for k, v in fn.args.items():
                    args[k] = v
                tool_calls.append(
                    JAImsToolCall(
                        id=fn.name,
                        tool_name=fn.name,
                        tool_args=args,
                    )
                )

            if txt := part.text:
                contents.append(txt)

            # TODO: Add image support

        return JAImsMessage(
            role=role,
            contents=contents,
            tool_calls=tool_calls,
            raw=response,
        )

    def __get_gemini_response(
        self,
        messages: List[JAImsMessage],
        tools: List[JAImsFunctionTool],
        stream: bool = False,
    ):

        # TODO: Add error and rate limit handling

        system_instruction, gemini_messages = self.__jaims_messages_to_gemini(messages)
        gemini_tools = self.__jaims_tools_to_gemini(tools) if tools else None

        multimodal_model = GenerativeModel(
            self.model,
            generation_config=self.generation_config,
            tools=gemini_tools,
            tool_config=self.tool_config,
            system_instruction=system_instruction or None,
        )

        response = multimodal_model.generate_content(
            contents=gemini_messages,
            stream=stream,
        )

        return response


def create_jaims_gemini(
    model: str = "gemini-1.5-pro",
    api_key: Optional[str] = None,
    generation_config: Optional[generation_types.GenerationConfigType] = None,
    tool_config: Optional[content_types.ToolConfigType] = None,
    history_manager: Optional[JAImsHistoryManager] = None,
    tool_manager: Optional[JAImsToolManager] = None,
    tools: Optional[List[JAImsFunctionTool]] = None,
) -> JAImsAgent:
    """
    Creates a JAIms instance with a Google Cloud AI adapter.

    Args:
        project_id (str): The Google Cloud project ID.
        location (str): The Google Cloud location.
        model (JAImsGeminiModel): The Gemini model to use. Defaults to JAImsGeminiModel.GEMINI_1_5_PRO.
        generation_config (Optional[GenerationConfig]): The generation configuration. Defaults to None.
        tool_config (Optional[ToolConfig]): The tool configuration. Defaults to None.
        history_manager (Optional[JAImsHistoryManager]): The history manager. Defaults to None.
        tool_manager (Optional[JAImsToolManager]): The tool manager. Defaults to None.
        tools (Optional[List[JAImsFunctionTool]]): The list of function tools. Defaults to None.

    Returns:
        JAImsAgent: The JAIms agent, initialized with the Google Cloud AI adapter.
    """

    adapter = JAImsGoogleGenerativeAIAdapter(
        api_key=api_key,
        model=model,
        generation_config=generation_config,
        tool_config=tool_config,
    )

    agent = JAImsAgent(
        llm_interface=adapter,
        history_manager=history_manager,
        tool_manager=tool_manager,
        tools=tools,
    )

    return agent



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .interfaces import JAImsLLMInterface, JAImsHistoryManager, JAImsToolManager

from typing import Generator, List, Optional


from jaims.default_tool_manager import JAImsDefaultToolManager


from jaims.entities import (
    JAImsMaxConsecutiveFunctionCallsExceeded,
    JAImsMessage,
    JAImsFunctionTool,
)

# TODO: Refactor all docstrings
# TODO: Refactor logging entirely
# TODO: Adjust imports in __init__.py files making them more explicit
# TODO: Adjust transaction storage for openai streaming response
# TODO: Add original response to jaims message and return it instead of strings
# TODO: Decide return type of runs, might not be ideal to return a string
# TODO: Implement a method to pass a function directly instead of the jaims descriptors
# TODO: Add images support

# TODO: Test Stuff:
# - Agent:
#  - with and without history
#  - with and without tools
#  - use mocks for the interfaces
# - History Managers []
# - Tool Manager []
#


class JAImsAgent:
    """
    Base  JAIms Agent class, interacts with the JAImsLLMInterface to run messages and tools.

    Args:
        llm_interface (JAImsLLMInterface): The LLM interface used for communication.
        history_manager (Optional[JAImsHistoryManager]): The history manager to track session messages. Defaults to None, which means no history is kept and messages are lost after run.
        tool_manager (Optional[JAImsToolManager]): The tool manager to handle tool calls. if None, the Default tool manager is used. Implement your own interface for advanced scenarios.
        tools (Optional[List[JAImsFunctionTool]]): The list of function tools. Defaults to None.

    Methods:
        run: Runs the agent with the given messages and tools.
        run_stream: Runs the agent in streaming mode with the given messages and tools.
    """

    def __init__(
        self,
        llm_interface: JAImsLLMInterface,
        history_manager: Optional[JAImsHistoryManager] = None,
        tool_manager: Optional[JAImsToolManager] = None,
        tools: Optional[List[JAImsFunctionTool]] = None,
    ):
        self.llm_interface = llm_interface
        self.tool_manager = tool_manager or JAImsDefaultToolManager()
        self.tools = tools or []
        self.history_manager = history_manager
        self.__session_iteration = -1
        self.__session_messages = []

    def __get_messages_for_run(
        self, messages: Optional[List[JAImsMessage]]
    ) -> List[JAImsMessage]:
        """
        Processes a list of messages for the agent run and, if necessary, adds them to history.

        Args:
            messages (Optional[List[JAImsMessage]]): The list of messages. Defaults to None.

        Returns:
            List[JAImsMessage]: The session messages.
        """
        session_messages = messages or []

        if self.history_manager:
            self.history_manager.add_messages(session_messages)
            session_messages = self.history_manager.get_messages()

        return session_messages

    def __update_session(
        self, session_messages: List[JAImsMessage], max_iterations: int
    ):
        self.__session_iteration += 1
        if self.__session_iteration > max_iterations:
            raise JAImsMaxConsecutiveFunctionCallsExceeded(self.__session_iteration)

        if self.history_manager:
            self.history_manager.add_messages(session_messages)
            self.__session_messages = self.history_manager.get_messages()
        else:
            self.__session_messages.extend(session_messages)

    def __end_session(self, response: Optional[JAImsMessage] = None):

        if self.history_manager and response:
            self.history_manager.add_messages([response])

        self.__session_iteration = -1
        self.__session_messages = []

    def __evaluate_tool_results(self, message: Optional[JAImsMessage]):
        tool_results = []
        if message and message.tool_calls:
            tool_call_results = self.tool_manager.handle_tool_calls(
                self, message.tool_calls, self.tools
            )
            tool_results = [message] + tool_call_results

        return tool_results

    def run(
        self,
        messages: Optional[List[JAImsMessage]] = None,
        max_iterations: int = 10,
    ) -> Optional[str]:
        """
        Runs the agent with the given messages and tools.

        Args:
            messages (Optional[List[JAImsMessage]]): The list of messages. Defaults to None.
            max_iterations (int): The maximum number of iterations. Defaults to 10. Pass this to constrain the maximum number of function calls.

        Returns:
            Optional[str]: The response text, or None if there is no response.
        """

        self.__update_session(messages or [], max_iterations)

        response_message = self.llm_interface.call(self.__session_messages, self.tools)

        tool_results = self.__evaluate_tool_results(response_message)
        if tool_results:
            return self.run(tool_results, max_iterations)

        self.__end_session(response_message)
        return response_message.get_text() or ""

    def run_stream(
        self,
        messages: Optional[List[JAImsMessage]] = None,
        max_iterations: int = 10,
    ) -> Generator[str, None, None]:
        """
        Runs the agent in streaming mode with the given messages and tools.

        Args:
            messages (Optional[List[JAImsMessage]]): The list of messages. Defaults to None.
            max_iterations (int): The maximum number of iterations. Defaults to 10. Pass this to constrain the maximum number of function calls.

        Yields:
            str: The text delta of each response message.
        """

        self.__update_session(messages or [], max_iterations)

        streaming_response = self.llm_interface.call_streaming(
            self.__session_messages, self.tools
        )

        response_message = None
        for delta_resp in streaming_response:
            response_message = delta_resp.message
            yield delta_resp.textDelta or ""

        tool_results = self.__evaluate_tool_results(response_message)
        if tool_results:
            yield from self.run_stream(tool_results, max_iterations)
            return

        self.__end_session(response_message)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .agent import JAImsAgent

from abc import ABC, abstractmethod
from typing import List, Generator


from jaims.entities import (
    JAImsMessage,
    JAImsStreamingMessage,
    JAImsToolCall,
    JAImsFunctionTool,
)


class JAImsLLMInterface(ABC):
    @abstractmethod
    def call(
        self, messages: List[JAImsMessage], tools: List[JAImsFunctionTool]
    ) -> JAImsMessage:
        raise NotImplementedError

    @abstractmethod
    def call_streaming(
        self, messages: List[JAImsMessage], tools: List[JAImsFunctionTool]
    ) -> Generator[JAImsStreamingMessage, None, None]:
        raise NotImplementedError


class JAImsHistoryOptimizer(ABC):
    @abstractmethod
    def optimize_history(self, messages: List[JAImsMessage]) -> List[JAImsMessage]:
        raise NotImplementedError


class JAImsHistoryManager(ABC):
    @abstractmethod
    def add_messages(self, messages: List[JAImsMessage]):
        raise NotImplementedError

    @abstractmethod
    def get_messages(self) -> List[JAImsMessage]:
        raise NotImplementedError


class JAImsToolManager(ABC):
    @abstractmethod
    def handle_tool_calls(
        self,
        agent: JAImsAgent,
        tool_calls: List[JAImsToolCall],
        tools: List[JAImsFunctionTool],
    ) -> List[JAImsMessage]:
        raise NotImplementedError



from typing import List, Optional

from jaims.entities import JAImsMessage
from jaims.interfaces import JAImsHistoryManager, JAImsHistoryOptimizer


class JAImsLastNHistoryOptimizer(JAImsHistoryOptimizer):
    def __init__(self, last_n: int):
        self.__last_n = last_n

    def optimize_history(self, messages: List[JAImsMessage]) -> List[JAImsMessage]:
        return messages[-self.__last_n :] if self.__last_n > 0 else []


class JAImsDefaultHistoryManager(JAImsHistoryManager):
    def __init__(
        self,
        history: Optional[List[JAImsMessage]] = None,
        leading_prompts: Optional[List[JAImsMessage]] = None,
        trailing_prompts: Optional[List[JAImsMessage]] = None,
        history_optimizer: Optional[JAImsHistoryOptimizer] = None,
    ):
        self.__history = history or []
        self.__leading_prompts = leading_prompts or []
        self.__trailing_prompts = trailing_prompts or []
        self.__history_optimizer = history_optimizer

    def add_messages(self, messages: List[JAImsMessage]):
        self.__history.extend(messages)

    def get_messages(self):

        if self.__history_optimizer:
            return (
                self.__leading_prompts
                + self.__history_optimizer.optimize_history(self.__history)
                + self.__trailing_prompts
            )

        return self.__leading_prompts + self.__history + self.__trailing_prompts

    def clear_history(self):
        """
        Clears the history.
        """
        self.__history = []

    def get_history(self):
        """
        Returns entire history.
        """
        return self.__history



from .adapter import (
    JAImsOpenaiKWArgs,
    JAImsOptions,
    create_jaims_openai,
    JAImsGPTModel,
    JAImsTokenHistoryOptimizer,
    OpenAITransactionStorageInterface,
    JAImsOpenaiAdapter,
)
